# SNP query tool for answering 15 questions used for a Bayesian analysis.
#    Copyright (C) <2012>  Thomas Freina and Christian X. Weichenberger
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# $Id: EnsemblAccess.pm 63 2012-04-10 12:58:27Z chris $
#
# Modifications:
# 2012-04-10 (CW) Special treatment in constrained elements

use strict;
use warnings;

package Eurac::EnsemblAccess;

use Carp;

use Bio::EnsEMBL::Registry;

use Eurac::SnpDataObject;

use constant DEBUG            => 1;
use constant STORE_DEBUG_DATA => 0;

our $ENSEMBL_INIT_DONE = 0;

# no args
sub ensemble_init
{
	return if $ENSEMBL_INIT_DONE;    # init already done

	print("\n") if DEBUG;

	print("TRACE INIT Eurac::EnsemblAccess\n") if DEBUG;

	my $registry = 'Bio::EnsEMBL::Registry';

	print("TRACE Ensembl load_registry_from_db()\n") if DEBUG;
	$registry->load_registry_from_db(
		-host => 'ensembldb.ensembl.org',
		-user => 'anonymous'
	);

	#print("TRACE Ensembl get_adaptor(Human,Core,Slice)\n") if DEBUG;
	#our $slice_adaptor = $registry->get_adaptor( 'Human', 'Core', 'Slice' );

	print("TRACE Ensembl get_adaptor(Human,Core,Gene)\n") if DEBUG;
	our $gene_adaptor = $registry->get_adaptor( 'Human', 'Core', 'Gene' );

	print("TRACE Ensembl get_adaptor(Human,Variation,Variation)\n") if DEBUG;
	our $va_adaptor = $registry->get_adaptor( 'Human', 'Variation', 'Variation' );

	print("TRACE Ensembl get_adaptor(Human,Variation,VariationFeature)\n") if DEBUG;
	our $vf_adaptor = $registry->get_adaptor( 'Human', 'Variation', 'VariationFeature' );

	print("TRACE Ensembl get_adaptor(Human,Funcgen,RegulatoryFeature)\n") if DEBUG;
	our $rf_adaptor = $registry->get_adaptor( 'Human', 'Funcgen', 'RegulatoryFeature' );

	print("TRACE Ensembl get_adaptor(Multi,Compara,MethodLinkSpeciesSet)\n") if DEBUG;
	our $mlss_adaptor = $registry->get_adaptor( 'Multi', 'Compara', 'MethodLinkSpeciesSet' );

	print("TRACE Ensembl get_adaptor(Multi,Compara,ConstrainedElement)\n") if DEBUG;
	our $ce_adaptor = $registry->get_adaptor( 'Multi', 'Compara', 'ConstrainedElement' );

	# fetch "constant" MethodLinkSpeciesSet only once
	my $methodlinktype = "GERP_CONSTRAINED_ELEMENT";    # not "PECAN";
	my $speciessetname = "amniotes";
	print( "TRACE FETCH Ensembl -> MethodLinkSpeciesSet <- fetch_by_method_link_type_species_set_name('",
		$methodlinktype, "','", $speciessetname, "')\n" )
	  if DEBUG;
	our $method_link_species_set =
	  $mlss_adaptor->fetch_by_method_link_type_species_set_name( $methodlinktype, $speciessetname );
	print( "Data/Ensembl: MethodLinkSpeciesSet '", $method_link_species_set->name(), "'\n" );

	print("\n") if DEBUG;

	$ENSEMBL_INIT_DONE = 1;
}

# arg1: string snpid
# return: SnpDataObject
sub fetch_snp_data_from_ensembl
{
	my $snpid = shift;

	# prepare the data-object with empty data-structures

	# digest data
	my %consequencetypes         = ();
	my %proteinfeature_hits      = ();
	my %regulatoryfeature_hits   = ();
	my @constrainedelements_hits = ();
	my @entrez_ids               = ();
	my %gene_0kb                 = ();
	my %gene_5kb                 = ();
	my %gene_5mb                 = ();

	# TODO change access in code to gene_0kb{'stable_ids'}
	my %gene_stable_ids = ();
	$gene_0kb{'stable_id_to_entrez'} = \%gene_stable_ids;

	my %debug_data = ();
	my $data       = Eurac::SnpDataObject->new(
		# digest of findings
		consequencetypes         => \%consequencetypes,
		proteinfeature_hits      => \%proteinfeature_hits,
		regulatoryfeature_hits   => \%regulatoryfeature_hits,
		constrainedelements_hits => \@constrainedelements_hits,
		entrez_ids               => \@entrez_ids,
		gene_0kb                 => \%gene_0kb,
		gene_5kb                 => \%gene_5kb,
		gene_5mb                 => \%gene_5mb,
		# raw data
		debug_data => \%debug_data
	);

	# do EnsEMBL init
	&ensemble_init();

	#get the Variation from the database using the name
	print( "TRACE FETCH Ensembl -> Variation <- fetch_by_name('", $snpid, "')\n" ) if DEBUG;
	my $variation = $Eurac::EnsemblAccess::va_adaptor->fetch_by_name($snpid);
	print( "Data/Ensembl: Variation '", $variation->source(), ":", $variation->name(), "'\n" );

	print("TRACE FETCH Ensembl -> VariationFeature[] <- fetch_all_by_Variation(...)\n") if DEBUG;
	my $variationfeatures_ref = $Eurac::EnsemblAccess::vf_adaptor->fetch_all_by_Variation($variation);
	my @variationfeatures     = @{$variationfeatures_ref};
	print( "Data/Ensembl: VariationFeature[], ", $#variationfeatures + 1, " element(s)\n" );

	# optional raw data, stored per slice
	my %constrainedelements_info = ();

	# a "cache" to avoid recurring fetches
	my @gene_checked_transcript_dbIDs = ();

	# this outer loop should only be run once. Most Ensembl example code with
	# snps does not loop over the variation features but simply takes the 
	# first (and only) entry of the array (ie list).
	# from the method description of Variation::fetch_all_by_Variation:
	# Description: Retrieves all variation features for a given variation. Most
	#              variations should only hit the genome once and only a return
	#              a single variation feature.

	foreach my $vf (@variationfeatures)
	{
		print( "TRACE Ensembl VariationFeature: ", $vf->start, "..", $vf->end, "\n" )
		  if DEBUG;    # TODO sure, theses are genomic coord?

		# do not take $vf->get_consequence_type() alone! go through TranscriptVariations, too.
		#foreach my $ct ( @{ $vf->consequence_type() } )
		#{
		#	print( "Data/Ensembl: Predicted variation consequence type '", $ct, "'\n" );
		#	push( @{ $consequencetypes{$ct} }, "(from VariationFeature)" );

			#push( @{ $data->consequencetypes{$ct} }, "(from VariationFeature)" );
		#}

		# regulatory regions
		# loop through all regulatory features. this could be usable later.
		# currently, we do not need these features. they are the Reg. Feats
		# grey blocks in the ensembl browser.
		for my $rfv ( @{$vf->get_all_RegulatoryFeatureVariations()} ) {
			print("TRACE/Ensembl: ", $rfv->feature->stable_id, ": ",
				  $rfv->display_consequence, ", '",$rfv->feature->feature_set->name,
				  "', range: ", $rfv->feature->start, "--", $rfv->feature->end,
				  "\n") if DEBUG;
			push( @{$regulatoryfeature_hits{$rfv->display_consequence}},
				  $rfv->feature->feature_set->name );
		}
		# loop through what Ensembl calls External Features. These are
		# from external databases. That's what we need.
		for my $erfv ( @{$vf->get_all_ExternalFeatureVariations()} ) {
			my $fname = $erfv->feature()->feature_set()->name();
			print("Data/Ensembl: in RegulatoryFeature found database hit \'$fname\'\n");
			if(    $fname eq "miRanda miRNA targets" ) {
				push( @{$regulatoryfeature_hits{"databases"}}, "miRanda" );
			}
			elsif( $fname eq "cisRED motifs" ) {
				push( @{$regulatoryfeature_hits{"databases"}}, "cisRED" );
			}
			elsif( $fname eq "VISTA enhancer set" ) {
				push( @{$regulatoryfeature_hits{"databases"}}, "VISTA" );
			}
		}

		# this is the slice associated with the SNP. highly useful, as it
		# leads us to genes for this SNP.
		my $snp_slice = $vf->feature_Slice();
		print( "Data/Ensembl: SNP Slice '", $snp_slice->display_id, "'\n" );

		# XXX this code has become quite ugly. ideally, a function to handle
		# each of these slice blocks is called.
		# get the gene associated with this transcript, $tv->transcript.
		my @genes = @{$snp_slice->get_all_Genes()};
		print( "Data/Ensembl: Number of genes associated with this SNP: ",
			   $#genes+1, " \n") if DEBUG;
		foreach my $gene (@genes)
		{
			print( "Data/Ensembl: Gene '", $gene->stable_id, "'\n" );
			print( "Data/Ensembl: gene_stable_id '", $gene->stable_id, "'\n" );
			# this table will be filled with crossreferences to other major
			# databases, eg. ncbi entrez.
			$gene_stable_ids{ $gene->stable_id } = ();

			# insert into list, for compliance with gene_5kb
			push( @{ $gene_0kb{'stable_id'} }, $gene->stable_id );
		}

		# now getting ±5kb genes
		my $slice5kb = $snp_slice->expand( 5000, 5000 );
		print( "Data/Ensembl: 5kb expanded Slice '", $slice5kb->display_id, "'\n" );

		# as an info for testing
		${ $data->get_gene_5kb() }{'slice_display_id'} = $slice5kb->display_id;

		print("TRACE Ensembl Genes[] <- get_all_Genes() [5Kb] \n") if DEBUG;
		@genes = @{ $slice5kb->get_all_Genes() };
		print( "Data/Ensembl: Genes[], ", $#genes + 1, " element(s)\n" );

		foreach my $gene (@genes)
		{
			# insert into list
			push( @{ $gene_5kb{'stable_id'} }, $gene->stable_id );
		}

		# now getting ±5Mb genes
		my $slice5mb = $snp_slice->expand( 5000000, 5000000 );
		print( "Data/Ensembl: 5Mb expanded Slice '", $slice5mb->display_id, "'\n" );

		# as an info for testing
		${ $data->get_gene_5mb() }{'slice_display_id'} = $slice5mb->display_id;

		print("TRACE Ensembl Genes[] <- get_all_Genes() [5Mb]\n") if DEBUG;
		@genes = @{ $slice5mb->get_all_Genes() };
		print( "Data/Ensembl: Genes[], ", $#genes + 1, " element(s)\n" );

		foreach my $gene (@genes)
		{
			# insert into list
			push( @{ $gene_5mb{'stable_id'} }, $gene->stable_id );
		}

		print("TRACE Ensembl TranscriptVariation[] <- get_all_TranscriptVariations()\n") if DEBUG;
		my $transcriptvariations_ref = $vf->get_all_TranscriptVariations();
		if ( defined $transcriptvariations_ref )
		{
			my @transcriptvariations = @{$transcriptvariations_ref};
			print( "Data/Ensembl: TranscriptVariations[], ", $#transcriptvariations + 1, " element(s)\n" );

			# This loops over all transcripts where this SNP is involved in.
			# I do not know which transcripts are taken, as even those outside
			# the 3' and 5' UTR regions are considered, if they are close enough.
			# eg. rsrs117884450, and transcript ENST00000427560 has consequence
			# type 'UPSTREAM', this is outside 5' UTR. UPSTREAM is defined as
			# 5kb, so this makes sense.
			foreach my $tv (@transcriptvariations)
			{
				# $tv->transcript_stable_id is the 'ENST...' identifier.
				print( "TRACE Ensembl TranscriptVariation: '", $tv->transcript_stable_id, "'\n" ) if DEBUG;

				# well, we are in a transcript variation. Therefore, there is
				# a transcript. However, a call to get_all_TranscriptVariations()
				# also returns transcripts in the VICINITY of this fucking snp.
				# eg.: rs13538 results in ENST00000272425 (NAT8 gene) and
				# ENST00000450720 (ALMS1P gene), but the SNP is on the NAT8 gene.
				# We therefore need to check if the snp overlaps with the
				# transcript.
				# First, transform the transcript to the chromosomal cs, as the
				# SNP is also in chrom. cs.
				my $snp_pos        = $vf->start();
				my $chr_transcript = $tv->transcript->transform("chromosome");
				# check for overlap, but first check if the transform went ok,
				# as in very rare cases this is not the case, eg. rs739439.
				if( defined($chr_transcript) &&
					($snp_pos<$chr_transcript->start() || $snp_pos>$chr_transcript->end()) ) 
				{
				  # none of the 15 questions concerns a region outside the 3'/5'
				  # UTR, except the regulatory questions and these are answered
				  # not on the basis of transcripts. These 5kb 5'/3' regions
				  # are termed UPSTREAM and DOWNSTREAM in Ensembl.
				  print("Data/Ensembl: Skipping transcript: SNP is outside of it.\n");
				  next;
				}

				# get the associated gene. This seems to be possible only with
				# the gene adaptor, not from the transcript itself.
				my $gene = $Eurac::EnsemblAccess::gene_adaptor->fetch_by_transcript_id($tv->transcript->dbID);
				print("Data/Ensembl: Gene associated with this TranscriptVariation: '", $gene->stable_id(), "'\n");

				# references to other data bases, we are interested in Entrez Gene ID.
				print("TRACE Ensembl DBEntry[] <- get_all_DBEntries()\n") if DEBUG;
				foreach my $dbe ( @{ $gene->get_all_DBEntries() } )
				{
					print( "TRACE Ensembl DBEntry dbname '", $dbe->dbname, "': '",
						 $dbe->primary_id, "'.\n") if DEBUG;
					if ( $dbe->dbname eq "EntrezGene" &&
						 not( $dbe->primary_id ~~@entrez_ids ) )
					{
						print( "Data/Ensembl: EntrezId '", $dbe->primary_id, "'\n" );
						push( @entrez_ids, $dbe->primary_id );
						push( @{ $gene_stable_ids{ $gene->stable_id } }, $dbe->display_id . "(" . $dbe->primary_id . ")" );
					}
				}

				# loop over all consequence types for this single transcript.
				# most of the time this will contain a single entry only.
				my $enst = $tv->transcript_stable_id;
				foreach my $ct ( @{ $tv->consequence_type() } )
				{
					print( "Data/Ensembl: Predicted variation consequence type '", $ct, "'\n" );
					push( @{ $consequencetypes{$ct} }, $enst );
				}

				# this should never happen, as we loop over all possible
				# transcripts of a single gene. (except for the super rare
				# cases where a genomic region holds multiple genes in a frame
				# shift. exists, but never seen. Francisco: virus genomes.)
				if ( $tv->transcript->dbID ~~@gene_checked_transcript_dbIDs )
				{
					print( "TRACE tv.transcript_id.dbID ", $tv->transcript->dbID, " already checked\n" ) if DEBUG;
					next;
				}

				# the transcript object can be retrieved from the
				# transcript variation object.
				my $transcript = $tv->transcript();
				my $snp_transcript_hit = 0;
				if ( $snp_pos >= $transcript->start && $snp_pos <= $transcript->end )
				{
					# insert into list of transcripts
					push( @{ $gene_0kb{'transcript'} }, $transcript->display_id . " " . $transcript->start . "..(" . $snp_pos . ").." . $transcript->end );
					$snp_transcript_hit = 1;
				}
				else
				{
					push( @{ $gene_0kb{'no_hit_transcript'} }, $transcript->display_id . " " . $transcript->start . ".." . $transcript->end . " / " . $snp_pos );
				}

				my $translation = $transcript->translation();
				if ( !defined $translation )
				{
					print("TRACE no translation\n") if DEBUG;

					# for debugging
					if ( not( $gene->stable_id ~~@{ $gene_0kb{'no_translation'} } ) )
					{
						push( @{ $gene_0kb{'no_translation'} }, $transcript->display_id );
					}

					next;
				}

				# check translation boundaries
				if ( $snp_pos >= $translation->genomic_start && $snp_pos <= $translation->genomic_end )
				{
					if( $snp_transcript_hit )
					{
						# insert into list of translations
						push( @{ $gene_0kb{'translation'} }, $transcript->display_id . " " . $translation->display_id . " " . $translation->genomic_start . "..(" . $snp_pos . ").." . $translation->genomic_end );
					}
					else
					{
						push( @{ $gene_0kb{'translation_of_no_hit_transcript'} }, $transcript->display_id . " " . $translation->display_id . " " . $translation->genomic_start . "..(" . $snp_pos . ").." . $translation->genomic_end );
					}
				}
				else
				{
					push( @{ $gene_0kb{'no_hit_translation'} }, $transcript->display_id . " " . $translation->display_id . " " . $translation->genomic_start . ".." . $translation->genomic_end . " / " . $snp_pos );
				}

				print("TRACE Ensembl ProteinFeatures[] <- get_all_ProteinFeatures()\n") if DEBUG;
				my $proteinfeatures_ref = $translation->get_all_ProteinFeatures();
				my @proteinfeatures     = @{$proteinfeatures_ref};
				foreach my $pf (@proteinfeatures)
				{
					my $logic_name = $pf->analysis()->logic_name();
					if ( $logic_name eq "pfam" )
					{
						my $start = $pf->start();
						my $end   = $pf->end();

						# projection or translation is nasty, as it needs a slice
						my $trmapper = Bio::EnsEMBL::TranscriptMapper->new($transcript);
						my @coords = $trmapper->pep2genomic( $start, $end );
						foreach my $coord (@coords)
						{

							#print("TRACE Ebl ProteinFeature: '",$pf->display_id, "' genomic coord ",$coord->start, "..", $coord->end, " \n") if DEBUG;
							# check if SNP coordinate (genomic) is within the interval
							if ( $snp_pos >= $coord->start && $snp_pos <= $coord->end )
							{
								my $id = $pf->display_id;    # TODO better id available?
								foreach my $tmpct ( @{ $tv->consequence_type() } )
								{
									if ( not( $id ~~@{ $proteinfeature_hits{$tmpct}{$logic_name} } ) )
									{
										print( "Data/Ensembl: in ProteinFeature '", $id, "' for ", $tmpct, "\n" );
										push( @{ $proteinfeature_hits{$tmpct}{$logic_name} }, $id );
									}
								}                            # foreach tmpct
							}
						}    # foreach pst
					}
				}    # foreach pf

				# getting SNP-slice
				my $slice = $vf->feature_Slice();

				# now processing Compara
				# using SNP-slice from above
				print("TRACE FETCH Ensembl -> ConstrainedElement[] <- fetch_all_by_MethodLinkSpeciesSet_Slice(...)\n")
				  if DEBUG;

				# rs10923010 reports a slice with start = end + 1, wtf?
				# this in turn corrupts a SQL statement formed in 
				# ConstrainedElementAdaptor::_fetch_all_ConstrainedElements
				my @constrainedelements = ();
				if( $slice->start<=$slice->end )
				{
					my $constrainedelements_ref =
						$Eurac::EnsemblAccess::ce_adaptor->fetch_all_by_MethodLinkSpeciesSet_Slice(
							$Eurac::EnsemblAccess::method_link_species_set, $slice );
					@constrainedelements = @{$constrainedelements_ref};
				}
				print( "Data/Ensembl: ConstrainedElement[], ", $#constrainedelements + 1, " element(s)\n" );

				foreach my $ce (@constrainedelements)
				{
					print( "SLICE ", $ce->slice()->display_id, "\n" );

					my $id = $ce->start . ".." . $ce->end;
					if ( not( $id ~~@constrainedelements_hits ) )
					{
						print( "Data/Ensembl: ConstrainedElement '",
							   $id, "' : tax ", $ce->taxonomic_level(), "\n" );
						push( @constrainedelements_hits, $id );
					}
				}

				# store fetched data for debugging
				$constrainedelements_info{ $slice->display_id } = \@constrainedelements;
				push( @gene_checked_transcript_dbIDs, $tv->transcript->dbID );
			}    # foreach tv
		}    # if def tv



	}    # foreach vf

	if (STORE_DEBUG_DATA)
	{

		# disconnect db ... TODO does not work as intended
		&disconnect($variation);
		foreach my $vf (@variationfeatures)
		{
			&disconnect($vf);
		}
		foreach my $k ( keys %constrainedelements_info )
		{
			foreach my $ce ( @{ $constrainedelements_info{$k} } )
			{
				&disconnect($ce);
			}
		}

		$debug_data{'ensembl_variation'}           = $variation;
		$debug_data{'ensembl_variationfeature'}    = \@variationfeatures;
		$debug_data{'ensembl_constrainedelements'} = \%constrainedelements_info;
	}

	return $data;
}

# arg1: Object to disconnect
sub disconnect
{
	my $self = shift;
	return unless ( exists $self->{'_adaptor'} );
	print("TRACE Ensembl disconnect()\n") if DEBUG;
	$self->{'_adaptor'}->dbc->db_handle->disconnect();
	delete $self->{'_adaptor'};
}

1;
