# SNP query tool for answering 15 questions used for a Bayesian analysis.
#    Copyright (C) <2012>  Thomas Freina and Christian X. Weichenberger
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# $Id: FileAccess.pm 59 2011-12-15 17:32:07Z chris $

use strict;
use warnings;

package Eurac::FileAccess;

use Carp;

use Eurac::SnpDataObject;

use constant DEBUG => 1;

use constant FILENAME_GENE_STUDY            => 'ensembl2huge.csv';
use constant FILENAME_GENE_PATHWAY          => 'ensembl2keggpathway.csv';
use constant FILENAME_PHENOTYPE_PATHWAY     => 'huge2keggpathway.csv';
use constant FILENAME_PHENOTYPE_PATHWAY_BL  => 'huge2keggpathway-blacklisted.csv';
use constant FILENAME_PHENOTYPE_PATHWAY_SBL => 'huge2keggpathway-shortblacklisted.csv';
# Currently, there is no difference between short and normal white lists,
# as they are generated by hand (Martin Goegele does this)
use constant FILENAME_PHENOTYPE_PATHWAY_WL  => 'huge2keggpathway-shortwhitelisted.csv';
use constant FILENAME_GENE_INTERACTION      => 'ens-binary-interactions.csv';
use constant FILENAME_PHENOTYPE_INTERACTION => 'huge2interacting-ensembl.csv';
use constant FILENAME_GENE_TISSUE           => 'ensembl2tissue.csv';
use constant FILENAME_OMIM_GENE             => 'mouse-pheno2ensembl.csv';

# arg1: string: snpid
# arg2: array of string: phenotypes of interest
# arg3: array of string: names of tissues of interest
# arg4: array of string: OMIM disease identifiers.
# arg5: string: data-files directory
# arg6: string. "white", "black", "shortblack", "none" to indicate the type
#       of exclusion list.
# arg7: SnpDataObject
sub fetch_snp_data_from_files
{
	my $snpid          = shift;
	my $phenotypes_ref = shift;
	my $tissues_ref    = shift;
	my $omims_ref      = shift;
	my $dataDirectory  = shift;
	my $useBWlist      = shift;
	my $data           = shift;

	Carp::croak("phenotypes not defined") if ( !defined $phenotypes_ref );

	my @phenotypes = @{$phenotypes_ref};
	my @tissues    = @{$tissues_ref};
	my @omims      = @{$omims_ref};
	Carp::croak("phenotypes empty") if ( $#phenotypes + 1 < 1 );
	Carp::carp("tissues empty") if ( $#tissues + 1 < 1 );
	Carp::carp("omims empty") if ( $#omims + 1 < 1 );

	if ( defined($data) )
	{
		Carp::croak("$data is not an Eurac::SnpDataObject object")
		  unless ( $data->isa("Eurac::SnpDataObject") );
	}
	else
	{
		Carp::croak("undefined Eurac::SnpDataObject object");
	}

	if ( !defined ${ $data->get_gene_5kb() }{'stable_id'} )
	{
		print("Data: no gene_5kb stable_ids\n");
		return;
	}

	my @genes5kb = @{ ${ $data->get_gene_5kb() }{'stable_id'} };

	my $gene5kb_data_ref = $data->get_gene_5kb();
	my $gene5mb_data_ref = $data->get_gene_5mb();

	&process_gene_study( \@phenotypes, $gene5kb_data_ref, $dataDirectory );
	&process_gene_study( \@phenotypes, $gene5mb_data_ref, $dataDirectory );

	&process_gene_pathway( \@phenotypes, \@genes5kb, $useBWlist, $dataDirectory, $data );

	&process_gene_interaction( \@phenotypes, \@genes5kb, $dataDirectory, $data );

	&process_gene_tissue_expression( \@tissues, \@genes5kb, $dataDirectory, $data );

	&process_phenotype_in_animal_models( \@omims, \@genes5kb, $dataDirectory, $data );
}

# local function
#
# arg1: array of string: phenotypes
# arg2: array of string: genes
# arg3: string: data-files directory
# arg4: reference to hash (part of SnpDataObject) for storing results (5kb, etc)
sub process_gene_study
{
	my $phenotypes_ref = shift;
	my $storage_ref    = shift;
	my $dataDirectory  = shift;

	my @phenotypes = @{$phenotypes_ref};
	my @genes      = @{$storage_ref->{'stable_id'}};

	my $pattern = "^(" . join( "|", @phenotypes ) . ")\\t(" . join( "|", @genes ) . ")\$";
	print( "TRACE PhenotypeGenePattern '", $pattern, "'\n" ) if DEBUG;

	my @lines = &grep_file( $pattern, $dataDirectory . '/' . FILENAME_GENE_STUDY );
	if ( $#lines + 1 < 1 )
	{
		print( "Data: no study match in ", FILENAME_GENE_STUDY, "\n" );
		push( @{$storage_ref->{'no_study_match'}}, FILENAME_GENE_STUDY );
		return;
	}
	print( "TRACE number of matches: ", $#lines + 1, "\n" ) if DEBUG;

	foreach my $line (@lines)
	{
		chomp($line);
		print( "TRACE Gene-Study '", $line, "'\n" ) if DEBUG;

		push( @{$storage_ref->{'study_match'}}, $line );
	}
}

# local function
#
# arg1: array of string: phenotypes
# arg2: array of string: genes
# arg3: string. type of black/white list.
# arg4: string: data-files directory
# arg5: SnpDataObject
sub process_gene_pathway
{
	my $phenotypes_ref = shift;
	my $genes_ref      = shift;
	my $whichListType  = shift;
	my $dataDirectory  = shift;
	my $data           = shift;

	my @phenotypes = @{$phenotypes_ref};
	my @genes      = @{$genes_ref};

	my %lst2fn = ("none"       => FILENAME_PHENOTYPE_PATHWAY,
				  "black"      => FILENAME_PHENOTYPE_PATHWAY_BL,
				  "shortblack" => FILENAME_PHENOTYPE_PATHWAY_SBL,
				  "white"      => FILENAME_PHENOTYPE_PATHWAY_WL);
	my $pwFileName = $lst2fn{$whichListType};

	my $g_pattern = "^(" . join( "|", @genes ) . ")";
	print( "TRACE GenePattern '", $g_pattern, "'\n" ) if DEBUG;

	my @gp_lines = &grep_file( $g_pattern, $dataDirectory . '/' . FILENAME_GENE_PATHWAY );
	if ( $#gp_lines + 1 < 1 )
	{
		print( "Data: no gene/pathway match in ", FILENAME_GENE_PATHWAY, "\n" );
		push( @{ ${ $data->get_gene_5kb() }{'no_pathway_match'} }, FILENAME_GENE_PATHWAY );
		return;
	}
	print( "TRACE number of matches: ", $#gp_lines + 1, "\n" ) if DEBUG;

	my @pathways;
	foreach my $line (@gp_lines)
	{
		chomp($line);
		my @parts = split( '\t', $line );
		push( @pathways, $parts[2] );
	}

	my $pattern = "^(" . join( "|", @phenotypes ) . ")\\t(" . join( "|", @pathways ) . ")\$";
	print( "TRACE PhenotypePathwayPattern '", $pattern, "'\n" ) if DEBUG;

	my @lines = &grep_file( $pattern, $dataDirectory . '/' . $pwFileName );
	if ( $#lines + 1 < 1 )
	{
		print( "Data: no cui/pathway match in ", $pwFileName, "\n" );
		push( @{ ${ $data->get_gene_5kb() }{'no_pathway_match'} }, $pwFileName );
		return;
	}
	print( "TRACE number of matches: ", $#lines + 1, "\n" ) if DEBUG;

	foreach my $line (@lines)
	{
		chomp($line);
		print( "TRACE Gene-Pathway '", $line, "'\n" ) if DEBUG;

		push( @{ ${ $data->get_gene_5kb() }{'pathway_match'} }, $line );
	}
}

# local function
#
# arg1: array of string: phenotypes
# arg2: array of string: genes
# arg3: string: data-files directory
# arg4: SnpDataObject
sub process_gene_interaction
{
	my $phenotypes_ref = shift;
	my $genes_ref      = shift;
	my $dataDirectory  = shift;
	my $data           = shift;

	my @phenotypes = @{$phenotypes_ref};
	my @genes      = @{$genes_ref};

	my $g_pattern = "^(" . join( "|", @genes ) . ")";
	print( "TRACE GenePattern '", $g_pattern, "'\n" ) if DEBUG;

	my @gi_lines = &grep_file( $g_pattern, $dataDirectory . '/' . FILENAME_GENE_INTERACTION );
	if ( $#gi_lines + 1 < 1 )
	{
		print( "Data: no interaction match in ", FILENAME_GENE_INTERACTION, "\n" );
		push( @{ ${ $data->get_gene_5kb() }{'no_interaction_match'} }, FILENAME_GENE_INTERACTION );
		return;
	}
	print( "TRACE number of matches: ", $#gi_lines + 1, "\n" ) if DEBUG;

	my @interactions;
	foreach my $line (@gi_lines)
	{
		chomp($line);
		my @parts = split( '\t', $line );
		push( @interactions, $parts[1] );
	}

	my $pattern = "^(" . join( "|", @phenotypes ) . ")\\t(" . join( "|", @interactions ) . ")\$";
	print( "TRACE PhenotypeInteractionsPattern '", $pattern, "'\n" ) if DEBUG;

	my @lines = &grep_file( $pattern, $dataDirectory . '/' . FILENAME_PHENOTYPE_INTERACTION );
	if ( $#lines + 1 < 1 )
	{
		print( "Data: no interaction match in ", FILENAME_PHENOTYPE_INTERACTION, "\n" );
		push( @{ ${ $data->get_gene_5kb() }{'no_interaction_match'} }, FILENAME_PHENOTYPE_INTERACTION );
		return;
	}
	print( "TRACE number of matches: ", $#lines + 1, "\n" ) if DEBUG;

	foreach my $line (@lines)
	{
		chomp($line);
		print( "TRACE Gene-Interaction '", $line, "'\n" ) if DEBUG;

		push( @{ ${ $data->get_gene_5kb() }{'interaction_match'} }, $line );
	}
}


# local function
#
# arg1: array of string: names of tissues of interest
# arg2: array of string: genes (ensembl IDs)
# arg3: string: data-files directory
# arg4: SnpDataObject
sub process_gene_tissue_expression
{
	my $tissues_ref    = shift;
	my $genes_ref      = shift;
	my $dataDirectory  = shift;
	my $data           = shift;

	# We allow the number of tissues to be 0. Take care of that.
	my @tissues = @{$tissues_ref};
	my @genes   = @{$genes_ref};

	my $pattern = "^(" . join( "|", @genes ) . ")\\t(" . join( "|", @tissues ) . ")\$";
	print( "TRACE GeneTissuePattern '", $pattern, "'\n" ) if DEBUG;

	my @lines = &grep_file( $pattern, $dataDirectory . '/' . FILENAME_GENE_TISSUE );
	if ( $#lines + 1 < 1 || $#tissues + 1 < 1 )
	{
		print( "Data: no tissue match in ", FILENAME_GENE_TISSUE, "\n" );
		push( @{ ${ $data->get_gene_5kb() }{'no_tissue_match'} }, FILENAME_GENE_TISSUE );
		return;
	}
	print( "TRACE number of matches: ", $#lines + 1, "\n" ) if DEBUG;

	foreach my $line (@lines)
	{
		chomp($line);
		print( "TRACE Gene-Tissue '", $line, "'\n" ) if DEBUG;

		push( @{ ${ $data->get_gene_5kb() }{'tissue_match'} }, $line );
	}
}

# local function
#
# arg1: reference to array of OMIM ids.
# arg2: reference to array of string: genes (ensembl IDs)
# arg3: string: data-files directory
# arg4: SnpDataObject
sub process_phenotype_in_animal_models
{
	my $omims_ref      = shift;
	my $genes_ref      = shift;
	my $dataDirectory  = shift;
	my $data           = shift;

	# number of entries in OMIMs can be 0. Take care of that.
	my @omims = @{$omims_ref};
	my @genes = @{$genes_ref};

	my $pattern = "^(" . join( "|", @omims ) . ")\\t(" . join( "|", @genes ) . ")\$";
	print( "TRACE OMIMGenePattern '", $pattern, "'\n" ) if DEBUG;

	my @lines = &grep_file( $pattern, $dataDirectory . '/' . FILENAME_OMIM_GENE );
	if ( $#lines + 1 < 1 || $#omims + 1 < 1 )
	{
		print( "Data: no OMIM match in ", FILENAME_OMIM_GENE, "\n" );
		push( @{ ${ $data->get_gene_5kb() }{'no_animal_model_match'} }, FILENAME_OMIM_GENE );
		return;
	}
	print( "TRACE number of matches: ", $#lines + 1, "\n" ) if DEBUG;

	foreach my $line (@lines)
	{
		chomp($line);
		print( "TRACE Gene-OMIM '", $line, "'\n" ) if DEBUG;

		push( @{ ${ $data->get_gene_5kb() }{'animal_model_match'} }, $line );
	}
}

# local function
#
# arg1: string pattern
# arg2: string filename
# return: array of string
sub grep_file
{
	my $pattern  = shift;
	my $filename = shift;

	open( my $fh, $filename )
	  or Carp::croak( "Could not open '", $filename, "': $!" );

	my @lines = grep( /$pattern/i, <$fh> );

	close($fh);

	return @lines;
}

1;
