# SNP query tool for answering 15 questions used for a Bayesian analysis.
#    Copyright (C) <2012>  Thomas Freina and Christian X. Weichenberger
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# $Id: Question.pm 54 2011-10-20 08:57:44Z chris $

use strict;
use warnings;

package Eurac::Question;

use Carp;

use Eurac::AnswerObject;
use Eurac::SnpDataObject;

# arg1: string: SNP-Id
# arg2: SnpDataObject
# return: AnswerObject
sub get_answers
{
	my $snpid    = shift;
	my $snp_data = shift;
	if ( defined($snp_data) )
	{
		Carp::confess("$snp_data is not an Eurac::SnpDataObject object")
		  unless ( $snp_data->isa("Eurac::SnpDataObject") );
	}
	else
	{
		Carp::confess("undefined Eurac::SnpDataObject object");
	}

	my @answers = ( ('?') x 15 );

	$answers[  1 - 1 ] = &get_answer01($snp_data);
	$answers[  2 - 1 ] = &get_answer02($snp_data);
	$answers[  3 - 1 ] = &get_answer03($snp_data);
	$answers[  4 - 1 ] = &get_answer04($snp_data);
	$answers[  5 - 1 ] = &get_answer05($snp_data);
	$answers[  6 - 1 ] = &get_answer06($snp_data);
	$answers[  7 - 1 ] = &get_answer07($snp_data);
	$answers[  8 - 1 ] = &get_answer08($snp_data);
	$answers[  9 - 1 ] = &get_answer09($snp_data);
	# TODO 10 not implemented ...
	$answers[ 11 - 1 ] = &get_answer11($snp_data);
	$answers[ 12 - 1 ] = &get_answer12($snp_data);
	$answers[ 13 - 1 ] = &get_answer13($snp_data);
	$answers[ 14 - 1 ] = &get_answer14($snp_data);
	$answers[ 15 - 1 ] = &get_answer15($snp_data);

	my $answerObject = Eurac::AnswerObject->new( snpid => $snpid, answers => \@answers );

	return $answerObject;
}

# 01: The SNP is in a transcribed but not translated region.
#
# arg1: SnpDataObject
# return: 'Y' or 'N'
sub get_answer01
{
	my $snp_data = shift;

	# Francisco: 'PARTIAL_CODON' is always coding.
	# rs6578323 (e63) offers basically everything: exon, 5' utr, 3' utr, 
	# first codon of an exon, last codon of an exon.
	# We skip the 'SPLICE_SITE' as it can be either in an intron or an exon.
	# Most of the time it occurs in combination with other, more informative 
	# terms which will lead to a decision beforehand.
	my @ct = keys %{ $snp_data->get_consequencetypes() };
	return 'N' if( 'STOP_GAINED'~~ @ct || 'STOP_LOST' ~~ @ct || 
				   'COMPLEX_INDEL' ~~ @ct || 'FRAMESHIFT_CODING' ~~ @ct ||
				   'NON_SYNONYMOUS_CODING' ~~ @ct || 'SYNONYMOUS_CODING' ~~ @ct ||
				   'PARTIAL_CODON' ~~ @ct );
	return 'Y' if ( 'INTRONIC' ~~ @ct || '5PRIME_UTR' ~~ @ct || '3PRIME_UTR' ~~ @ct ||
					'WITHIN_NON_CODING_GENE' ~~ @ct || 'WITHIN_MATURE_miRNA' ~~ @ct ||
				    'ESSENTIAL_SPLICE_SITE' ~~ @ct );
	return 'N';
}

# 02: The SNP is in a translated genomic region (exon) but does not change the amino acid.
#
# arg1: SnpDataObject
# return: 'Y' or 'N'
sub get_answer02
{
	my $snp_data = shift;
	my @ct       = keys %{ $snp_data->get_consequencetypes() };
	return 'Y' if ( 'SYNONYMOUS_CODING' ~~@ct );
	return 'N';
}

# 03: The SNP changes the amino acid but is not in a functional protein domain.
#
# arg1: SnpDataObject
# return: 'Y' or 'N'
sub get_answer03
{
	my $snp_data = shift;
	my @ct       = keys %{ $snp_data->get_consequencetypes() };
	if (   'STOP_GAINED' ~~@ct
		|| 'STOP_LOST' ~~@ct
		|| 'FRAMESHIFT_CODING' ~~@ct
		|| 'NON_SYNONYMOUS_CODING' ~~@ct
		|| 'PARTIAL_CODON' ~~@ct
	    || 'COMPLEX_INDEL' ~~@ct )
	{
		my %pf = %{ $snp_data->get_proteinfeature_hits() };
		return 'Y'
		  if ( not( exists( $pf{'STOP_GAINED'}{'pfam'} ) )
			&& not( exists( $pf{'STOP_LOST'}{'pfam'} ) )
			&& not( exists( $pf{'FRAMESHIFT_CODING'}{'pfam'} ) )
			&& not( exists( $pf{'NON_SYNONYMOUS_CODING'}{'pfam'} ) )
			&& not( exists( $pf{'PARTIAL_CODON'}{'pfam'} ) )
			&& not( exists( $pf{'COMPLEX_INDEL'}{'pfam'} ) ) );
	}
	return 'N';
}

# 04: The SNP is in a functional protein domain.
#
# arg1: SnpDataObject
# return: 'Y' or 'N'
sub get_answer04
{
	my $snp_data = shift;
	my %pf       = %{ $snp_data->get_proteinfeature_hits() };
	my $count    = 0;

	foreach my $key ( keys %pf )
	{
		if ( exists( $pf{$key}{'pfam'} ) )
		{
			$count++;
		}
	}

	return 'Y' if ( $count > 0 );
	return 'N';
}

# Helper function:
# arg 1: reference to consequence type names array.
# return: 1 if the consequence type implies transcription.
#         0 otherwise.
sub is_transcribed
{
	my $consequence_types_ref = shift;
	my @ct                    = @$consequence_types_ref;
	return 1
	  if( 'ESSENTIAL_SPLICE_SITE' ~~ @ct || 'STOP_GAINED' ~~ @ct ||
		  'STOP_LOST' ~~ @ct || 'COMPLEX_INDEL' ~~ @ct || 
		  'FRAMESHIFT_CODING' ~~ @ct || 'NON_SYNONYMOUS_CODING' ~~ @ct ||
		  'SPLICE_SITE' ~~ @ct || 'PARTIAL_CODON' ~~ @ct ||
		  'SYNONYMOUS_CODING' ~~ @ct || 'REGULATORY_REGION' ~~ @ct ||
		  'WITHIN_MATURE_miRNA' ~~ @ct || '5PRIME_UTR' ~~ @ct ||
		  '3PRIME_UTR' ~~ @ct || 'INTRONIC' ~~ @ct ||
		  'NMD_TRANSCRIPT' ~~ @ct || 'WITHIN_NON_CODING_GENE' ~~ @ct );
	# the SNP is not in a transcribed region,
	# ie, upstream/downstream, or intergenic, or some HGMD mutation
	# database entry. this however is not in the consequence type list,
	# as we store them by looping over the transcripts!
	return 0;
}

# 05: The SNP is in a regulatory region which is not transcribed.
#
# arg1: SnpDataObject
# return: 'Y' or 'N'
sub get_answer05
{
	# we have two ways of defining transcription regulation.
	# a) by databases (cisRED, VISTA, miranda)
	# b) by the regulatory feature track.
	my $snp_data = shift;
	my %rf       = %{ $snp_data->get_regulatoryfeature_hits() };
	my @ct       = keys %{ $snp_data->get_consequencetypes() };

	# our definition is that these two databases below are describing features
	# on DNA and therefore are not transcribed. this can be erroneous.
	return 'Y'
	  if( 'cisRED' ~~ $rf{'databases'} || 'VISTA' ~~ $rf{'databases'} );

	return 'Y'
	  if( exists($rf{'REGULATORY_REGION'}) && !is_transcribed(\@ct) );

	return 'N';
}

# 06: The SNP is in a transcribed regulatory region.
#
# arg1: SnpDataObject
# return: 'Y' or 'N'
sub get_answer06
{
	my $snp_data = shift;
	my %rf       = %{ $snp_data->get_regulatoryfeature_hits() };
	my @ct       = keys %{ $snp_data->get_consequencetypes() };

	# see answer05.
	return 'Y' if ( 'miRanda' ~~ $rf{'databases'}  );

	# YES, if there is any sign of transcription.
	return 'Y'
	  if( exists($rf{'REGULATORY_REGION'}) && is_transcribed(\@ct) );

	return 'N';
}

# 07: The SNP is in a genomic region evolutionary conserved in vertebrates.
#
# arg1: SnpDataObject
# return: 'Y' or 'N'
sub get_answer07
{
	my $snp_data = shift;
	my @ce       = @{ $snp_data->get_constrainedelements_hits() };

	return 'Y' if ( $#ce + 1 > 0 );
	return 'N';
}

# 08. The SNP is in a gene (±5kb) that has been associated with the same/closely related phenotype in a meta-analysis or in more than one study.
#
# arg1: SnpDataObject
# return: 'X' or 'N'
sub get_answer08
{
	my $snp_data = shift;

	my %gene_5kb = %{ $snp_data->get_gene_5kb() };
	return 'X' if ( $#{ $gene_5kb{'study_match'} } + 1 > 0 );
	return 'N';
}

# 09. The SNP is in a gene (±5kb) that has been associated with the same/closely related phenotype in a single study.
#
# arg1: SnpDataObject
# return: 'X' or 'N'
sub get_answer09
{
	my $snp_data = shift;

	my %gene_5kb = %{ $snp_data->get_gene_5kb() };
	return 'X' if ( $#{ $gene_5kb{'study_match'} } + 1 > 0 );
	return 'N';
}

# 10. The SNP is in a gene (±5kb) which is under a linkage peak that has been associated with the same/closely related phenotype.

# 11. The SNP is in a locus within which other SNPs have been associated with the same/closely related phenotype.
# exchanged 'locus' to '+/- 5mb' and 'SNPs' to 'genes'
sub get_answer11
{
	my $snp_data = shift;

	my %gene_5mb = %{ $snp_data->get_gene_5mb() };
	return 'Y' if ( $#{ $gene_5mb{'study_match'} } + 1 > 0 );
	return 'N';
}


# 12. The SNP is in a gene (±5kb) that has been associated with the same/closely related phenotype in functional models (animal or in-vitro studies).
sub get_answer12
{
	my $snp_data = shift;

	my %gene_5kb = %{ $snp_data->get_gene_5kb() };
	return 'Y' if ( $#{ $gene_5kb{'animal_model_match'} } + 1 > 0 );
	return 'N';
}

# 13. The SNP is in a gene (±5kb) which is highly expressed in a tissue relevant to the phenotype.

sub get_answer13
{
	my $snp_data = shift;

	my %gene_5kb = %{ $snp_data->get_gene_5kb() };
	return 'Y' if ( $#{ $gene_5kb{'tissue_match'} } + 1 > 0 );
	return 'N';
}

# 14. The SNP is in a gene (±5kb) which encodes for a protein in a pathway relevant to the phenotype.
#
# arg1: SnpDataObject
# return: 'Y' or 'N'
sub get_answer14
{
	my $snp_data = shift;

	my %gene_5kb = %{ $snp_data->get_gene_5kb() };
	return 'Y' if ( $#{ $gene_5kb{'pathway_match'} } + 1 > 0 );
	return 'N';
}

# 15. The SNP is in a gene (±5kb) which shows gene/protein interactions relevant to the phenotype.
#
# arg1: SnpDataObject
# return: 'Y' or 'N'
sub get_answer15
{
	my $snp_data = shift;

	my %gene_5kb = %{ $snp_data->get_gene_5kb() };
	return 'Y' if ( $#{ $gene_5kb{'interaction_match'} } + 1 > 0 );
	return 'N';
}

1;
