#!/usr/bin/perl

# SNP query tool for answering 15 questions used for a Bayesian analysis.
#    Copyright (C) <2012>  Thomas Freina and Christian X. Weichenberger
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# call: ./snp_info.pl [options] snpid-file

use strict;
use warnings;

use Carp;
use Data::Dumper;
use Getopt::Std;

use Eurac::EnsemblAccess;
use Eurac::Diagnostics;
use Eurac::FileAccess;
use Eurac::InputParser;
use Eurac::Persistence;
use Eurac::Question;

# make STDOUT unbuffered
select STDOUT;
$| = 1;

# config Dumper
$Data::Dumper::Sortkeys = 1;
$Data::Dumper::Useqq    = 1;

use constant DEBUG        => 0;
use constant DEBUG_DUMP   => 0;

# These are from the BioGPS.py file and originally come from the BioGPS file
# headers.
my @TISSUETYPES = ("721_B_LYMPHOBLASTS", "ADIPOCYTE", "ADRENALCORTEX", "ADRENALGLAND",
"AMYGDALA", "APPENDIX", "ATRIOVENTRICULARNODE", "BDCA4+_DENTRITICCELLS",
"BONEMARROW", "BRONCHIALEPITHELIALCELLS", "CD105+_ENDOTHELIAL",
"CD14+_MONOCYTES", "CD19+_BCELLS(NEG._SEL.)", "CD33+_MYELOID", "CD34+",
"CD4+_TCELLS", "CD56+_NKCELLS", "CD71+_EARLYERYTHROID", "CD8+_TCELLS",
"CARDIACMYOCYTES", "CAUDATENUCLEUS", "CEREBELLUM", "CEREBELLUMPEDUNCLES",
"CILIARYGANGLION", "CINGULATECORTEX", "COLORECTALADENOCARCINOMA",
"DORSALROOTGANGLION", "FETALTHYROID", "FETALBRAIN", "FETALLIVER", "FETALLUNG",
"GLOBUSPALLIDUS", "HEART", "HYPOTHALAMUS", "KIDNEY",
"LEUKEMIA_CHRONICMYELOGENOUSK-562", "LEUKEMIA_PROMYELOCYTIC-HL-60",
"LEUKEMIALYMPHOBLASTIC(MOLT-4)", "LIVER", "LUNG", "LYMPHNODE",
"LYMPHOMA_BURKITTS(DAUDI)", "LYMPHOMA_BURKITTS(RAJI)", "MEDULLAOBLONGATA",
"OCCIPITALLOBE", "OLFACTORYBULB", "OVARY", "PANCREAS", "PANCREATICISLET",
"PARIETALLOBE", "PITUITARY", "PLACENTA", "PONS", "PREFRONTALCORTEX",
"PROSTATE", "SALIVARYGLAND", "SKELETALMUSCLE", "SKIN", "SMOOTHMUSCLE",
"SPINALCORD", "SUBTHALAMICNUCLEUS", "SUPERIORCERVICALGANGLION",
"TEMPORALLOBE", "TESTIS", "TESTISGERMCELL", "TESTISINTERSITIAL",
"TESTISLEYDIGCELL", "TESTISSEMINIFEROUSTUBULE", "THALAMUS", "THYMUS",
"THYROID", "TONGUE", "TONSIL", "TRACHEA", "TRIGEMINALGANGLION", "UTERUS",
"UTERUSCORPUS", "WHOLEBLOOD", "WHOLEBRAIN", "COLON", "PINEAL_DAY",
"PINEAL_NIGHT", "RETINA", "SMALL_INTESTINE");

Eurac::Diagnostics::print_INC() if DEBUG;

Eurac::Diagnostics::print_ARGV() if DEBUG;

my %options = &parse_argv();

Eurac::Diagnostics::print_ARGV() if DEBUG;

Eurac::Diagnostics::print_options( \%options ) if DEBUG;

Eurac::Diagnostics::print_ensembl_version();    # if DEBUG;

if( $options{v} )
{
	exit 0;
}

# phase 1: read and check the input

my @phenotypes;
my @tissues;
my @snpids;
my @omims;
&parse_input_lines( \@phenotypes, \@tissues, \@omims, \@snpids );

foreach my $snpid (@snpids)
{
	print( "TRACE SNP-Id: ", $snpid, " to be checked (again)\n" ) if DEBUG;

	if ( Eurac::InputParser::check_snpid($snpid) )
	{
		if ($snpid)
		{
			print( "TRACE SNP-Id: ", $snpid, " to be processed\n" ) if DEBUG;

			# phase 2: retrieve all data needed

			my $snp_data = undef;

			if ( $options{'c'} ) # if cache-dir defined, try to load SNP-data from dump-file
			{
				$snp_data = Eurac::Persistence::load_snp_data_object( $snpid, $options{'c'} );
				if( $options{'g'} )
				{
					write_5k_gene_names($snpid, $snp_data, $options{o});
					# whatever additional option the user has chosen, we skip it
					next;
				}
			}

			if ( !defined $snp_data )
			{
				$snp_data = Eurac::EnsemblAccess::fetch_snp_data_from_ensembl($snpid);
				print( "Data: ", Dumper($snp_data), "\n" ) if DEBUG_DUMP;

				my $listType="none";
				if( $options{'b'} )
				{
					$listType = "black";
					if( $options{'s'} )
					{
						$listType = "shortblack";
					}
				}
				if( $options{'w'} )
				{
					$listType = "white"
				}
				Eurac::FileAccess::fetch_snp_data_from_files( $snpid, \@phenotypes, \@tissues, \@omims, $options{'d'}, $listType, $snp_data );
				print( "Data: ", Dumper($snp_data), "\n" ) if DEBUG_DUMP;

				if ( $options{'c'} ) # if cache-dir defined, save SNP-data as dump-file
				{
					Eurac::Persistence::store_snp_data_object( $snpid, $options{'c'}, $snp_data );
				}
			}
#			# TEST TEST TEST
#			else
#			{
#				Eurac::FileAccess::fetch_snp_data_from_files( $snpid, \@phenotypes, $options{'d'}, $snp_data );
#			}
#			# TEST TEST TEST

			# phase 3: answer the questions

			my $answer_data = Eurac::Question::get_answers( $snpid, $snp_data );
			print( "Answers: ", Dumper($answer_data), "\n" ) if DEBUG_DUMP;

			# phase 4: produce output

			&write_answers( $snp_data, $answer_data, $options{'o'} );

			&print_answers( $snp_data, $answer_data );
			print("\n");
		}
		else
		{
			Carp::cluck( "Warning: SNP-Id \"", $snpid, "\" empty.\n" );
		}
	}
	else
	{
		Carp::cluck( "Warning: SNP-Id \"", $snpid, "\" invalid.\n" );
	}
}

# local function
#
# return: hash of string: Options
sub parse_argv
{
	my $default_data_dir    = "../data/";
	my $default_answer_file = "answer.csv";
	my %options = ( 'd' => $default_data_dir, 'o' => $default_answer_file, 
					'v' => 0, 'c' => '', 'g' => 0, 's' => 0, 'b' => 0, 
				'-w' => 0);
	getopt( 'doc', \%options );    # 'd', 'o' and 'c' are expecting another arg

	if ( $#ARGV + 1 != 1 && ! $options{v} || $options{'w'} && $options{'b'} )
	{
		print("Usage: snp_info [options] filename\n");
		print("\n");
		print("options:\n");
		print("  -d directory ... location of data files (default: '$default_data_dir')\n");
		print("  -o filename  ... file name of results (default: '$default_answer_file')\n");
		print("                   NOTE: output will be appended to any existing file.\n");
		print("\n");
		print("  -v           ... Print Ensembl version and quit.\n");
		print("  -c directory ... Perl dump files (default: empty, no cache)\n");
		print("                   if not empty: used as a cache, and for debugging.\n");
		print("  -g           ... Print gene names for all snps previously computed and stored\n");
		print("                   in a cache directory (requires -c option)\n");
		print("  -b           ... Use blacklists for question 14 (DEPRECATED).\n");
		print("  -w           ... Use whitelists for question 14 (DEPRECATED).\n");
		print("  -s           ... Use short blacklists for question 14. This is for running\n");
		print("                   the program on the input set with the smaller number of CUI\n");
		print("                   that were needed for a sensitiviy analysis.\n");
		print("                   (DEPRECATED)\n");
		print("\n");
		print("filename ... List of phenotypes, tissue names, and SNP Ids.\n");
		print("             Phenotype lines start with 'C'\n");
		print("             Tissue name lines start with 'T:'\n");
		print("             SNP-Id lines start with 'rs'\n");
		print("             Data entries can be mixed, there is no need for any\n");
		print("             order. Furthermore, entries are regarded case insensitive.\n");
		print("Options '-b' (black list) and '-w' (white list) are mutually exclusive!\n");
		print("Option '-s' (short list) is effective only with option '-b'\n");
		exit;
	}

	return %options;
}

# local function
# read and analyse the input
#
# arg1: empty array: for Phenotypes, list of strings
# arg2: empty array: for SNP-Ids, list of strings
sub parse_input_lines
{
	my $phenotypes = shift;
	my $tissues    = shift;
	my $omims      = shift;
	my $snpids     = shift;

	while (<>)
	{
		my $line = $_;

		chomp($line);    # avoid \n at the end of $line

		print( "TRACE line: '", $line, "'\n" ) if DEBUG;

		if ( Eurac::InputParser::is_comment($line) )
		{
			# NOP
		}
		elsif ( Eurac::InputParser::is_phenotype_line($line) )
		{
			my $phenotype = Eurac::InputParser::get_phenotype_from_phenotype_line($line);
			print( "TRACE Parsed Phenotype: ", $phenotype, "\n" ) if DEBUG;
			push( @{$phenotypes}, $phenotype );
		}
		elsif ( Eurac::InputParser::is_tissue_line($line) )
		{
			my $tissue = Eurac::InputParser::get_tissue_from_tissue_line($line);
			if( substr(uc($tissue),2) ~~ @TISSUETYPES )
			{
				print( "TRACE Parsed Tissue: $tissue\n" ) if DEBUG;
				push( @$tissues, $tissue );
			}
			else
			{
				Carp::croak("Unknown tissue ``$tissue'' in input file");
			}
		}
		elsif ( Eurac::InputParser::is_omim_line($line) )
		{
			my $omim = Eurac::InputParser::get_omim_from_omim_line($line);
			print( "TRACE Parsed OMIM ID: $omim\n " ) if DEBUG;
			push( @$omims, $omim );
		}
		elsif ( Eurac::InputParser::is_snp_line($line) )
		{
			my $snpid = Eurac::InputParser::get_snpid_from_snp_line($line);
			if ( Eurac::InputParser::check_snpid($snpid) )
			{
				print( "TRACE Parsed SNP-Id: ", $snpid, "\n" ) if DEBUG;
				push( @{$snpids}, $snpid );
			}
		}
		else
		{
			Carp::carp( "Invalid input, Input-Line '", $line, "' ignored.\n" );
		}
	}
}

# local function
#
# arg1: SnpDataObject
# arg2: AnswerObject
# arg3: string: filename to append answers
sub write_answers
{
	my $snp_data     = shift;
	my $answerObject = shift;
	my $fileName     = shift;

	# append to result-file
	open FILE, ">>$fileName" or Carp::croak( "Can\'t open file '", $fileName, "' for writing: $!\n" );

#	print( FILE $answerObject->get_snpid(),
#		";", join( ";", @{ $answerObject->get_answers() } ),
#		";", $snp_data->get_obj_timestamp(), ";\n"
#	);
	print( FILE $answerObject->get_snpid(),
		",", join( ",", @{ $answerObject->get_answers() } ), "\n"
	);

	# close result-file
	close FILE or Carp::croak( "Error closing '", $fileName, "': $!\n" );
}

sub write_5k_gene_names
{
	my $snpid    = shift;
	my $snp_data = shift;
	my $fileName = shift;
	my %gene_5kb = %{ $snp_data->get_gene_5kb() };

	# append to result-file
	open FILE, ">>$fileName" or Carp::croak( "Can\'t open file '", $fileName, "' for writing: $!\n" );

	foreach my $ens_gene_id (@{$gene_5kb{'stable_id'}})
	{
		printf(FILE "%s\t%s\n", $snpid, $ens_gene_id);
	}

	# close result-file
	close FILE or Carp::croak( "Error closing '", $fileName, "': $!\n" );
}


# local function
#
# arg1: SnpDataObject
# arg2: AnswerObject
sub print_answers
{
	my $snp_data     = shift;
	my $answerObject = shift;

	print( $answerObject->get_snpid(), ": ", join( " | ", @{ $answerObject->get_answers() } ), " |\n" );
	print("\n") if DEBUG;
}

