#!/bin/bash
# unittest.sh
# 
# SNP query tool for answering 15 questions used for a Bayesian analysis.
#    Copyright (C) <2012>  Thomas Freina and Christian X. Weichenberger
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Each of the 15 Bayesian questions is devoted at minimum one specific
# unit test. 
#
# Author       : Chris Weichenberger
# Created      : 2011-10-20
# Modifications:
# 2012-04-10 (CW) Removed unit test for black/white lists.
#
# $Id: unittest.sh 67 2012-04-11 10:35:56Z chris $

# Ensembl version for which this test has been written.
ENSVER="63"
# Executable for data retrieval.
PROG=snp_info.pl
# Options (they will be set dynamically depending on the type of test)
OPTS=""
# Data directory.
DATA=../data/
# Data for unit test input and comparison.
UTDATA=$DATA/unittest
# Output file name for temporary answer files.
OFN=/tmp/unittest-out.csv

# Run a single unit test.
# Runs the query program and stores the output in a file. This is then
# compared to the file with the data we are expecting. If there is a
# difference, the program will print an error message and quit the test with
# an exit code 1.
#
# arg 1: Input file name with cui, omim, tissue, and rs data.
# arg 2: Comparison file with expected results from this run.
function runtest {
   inputfn=$1
   comparefn=$2
   rm -f $OFN
   perl $PROG $OPTS -d $DATA -o $OFN $inputfn
   diff $OFN $comparefn
   result=$?
   if [ $result -ne 0 ]
   then
      echo "Unit test failed for input file \`\`$inputfn''.";
      exit 1;
   else
      echo "Unit test passed input file \`\`$inputfn''.";
   fi
}

# Sanity check. It is easy to upgrade to a new Ensembl version, but it is also
# pretty sure it will break this code in subtle ways, as people redefine some
# output and some data types from release to release. We have already observed
# this from v61 to v62 and to v63. Fortunately, old versions of Ensembl can
# still be used on the web and with the API.
thisver=`perl $PROG -v | awk '{print $3;}'`
if [ $thisver != $ENSVER ] 
then
   echo "Unit test has been written for Ensembl v$ENSVER. You are using v$thisver. Exiting."
   exit 1;
fi

# 2012-04-10
# Removed UT for black and white lists. This is no longer supported 
# anyway and has proven to be no big difference.
# Test the different options that can now be provided for lists.
# white list
#OPTS="-w"
#runtest $UTDATA/ut-options-whitelist-input.txt $UTDATA/ut-options-whitelist-answers.txt
# short black list
#OPTS="-b -s"
#runtest $UTDATA/ut-options-shortblacklist-input.txt $UTDATA/ut-options-shortblacklist-answers.txt
# black list
#OPTS="-b"
#runtest $UTDATA/ut-options-blacklist-input.txt $UTDATA/ut-options-blacklist-answers.txt

# from now on, no options for the SNP info program.
OPTS=""

# Run a unit test on each of the 15 questions from the Bayesian.
for i in `seq -f '%02.0f' 15`
do 
   runtest $UTDATA/ut-q$i-input.txt $UTDATA/ut-q$i-answers.txt
done

# This is a huge test with data from Morbus Crohn previously used in our
# project. 
runtest $UTDATA/ut-rndsel-input.txt $UTDATA/ut-rndsel-answers.txt

echo "Congrats. Unit test has successfully been completed. Press any key to delete"
echo "temporary files and finish this test. You may press <ctrl>-c now to save "
echo "\`\`$OFN'' from being deleted."
read
rm -f $OFN
